"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deepClone = void 0;
function deepClone(obj, parentify, allowedTypes, preserveTypes, precache) {
    console.log('deepclone', obj);
    const existing = precache !== null && precache !== void 0 ? precache : new Map();
    const result = deepCloneRecursive(obj, existing, parentify, allowedTypes, preserveTypes);
    return {
        result,
        cache: existing,
    };
}
exports.deepClone = deepClone;
function deepCloneRecursive(obj, existingClones, parentify, allowedTypes, preserveTypes) {
    const objClone = preserveTypes ? Object.create(Object.getPrototypeOf(obj)) : {};
    if (!existingClones.has(obj)) {
        existingClones.set(obj, objClone);
    }
    Object.assign(objClone, { '$original': obj });
    for (const prop in obj) {
        if (['$parents', '$original'].includes(prop)) {
            continue;
        }
        const propValue = obj[prop];
        // const t = typeof propValue;
        if ((typeof propValue === "object") && (!allowedTypes || allowedTypes.some(at => propValue instanceof at))) {
            let propValueClone;
            if (existingClones.has(propValue)) {
                propValueClone = existingClones.get(propValue);
                Object.assign(objClone, { [prop]: propValueClone });
            }
            else {
                propValueClone = deepCloneRecursive(propValue, existingClones, parentify, allowedTypes, preserveTypes);
                // // This is too late. Must be called before deepCloneRecursive
                // existingClones.set(propValue, clone);
                Object.assign(objClone, { [prop]: propValueClone });
            }
            if (parentify && propValueClone) {
                if (propValueClone['$parents']) {
                    Object.assign(propValueClone, { '$parents': [...propValueClone['$parents'], objClone] });
                }
                else {
                    Object.assign(propValueClone, { '$parents': [objClone] });
                }
            }
        }
        else if (propValue instanceof Array) {
            const arrayClones = propValue.map(v => deepCloneRecursive(v, existingClones, parentify, allowedTypes, preserveTypes));
            if (parentify) {
                arrayClones.filter(clone => clone).forEach(clone => {
                    if (clone['$parents']) {
                        Object.assign(clone, { '$parents': [...clone['$parents'], objClone] });
                    }
                    else {
                        Object.assign(clone, { '$parents': [objClone] });
                    }
                });
            }
            Object.assign(objClone, { [prop]: arrayClones });
        }
        else {
            Object.assign(objClone, { [prop]: propValue });
        }
    }
    return objClone;
}
//# sourceMappingURL=parentify.service.js.map